@@ -1,14 +1,16 @@
-import json
-import logging
-from _socket import gethostname
-from dataclasses import dataclass, fields
-from distutils.util import strtobool
 from os import getenv
-from pathlib import Path
 
+import logging
 import yaml
+from _socket import gethostname
 from appdirs import AppDirs
+from copy import deepcopy
+from dataclasses import dataclass, fields, field
+from distutils.util import strtobool
+from functools import lru_cache
 from getmac import getmac
+from pathlib import Path
+from typing import Dict, Optional
 
 NAME = 'amniotic'
 ORG = 'frontmatter'
@@ -16,7 +18,7 @@ APP_DIRS = AppDirs(NAME, ORG)
 MAC_ADDRESS = getmac.get_mac_address().replace(':', '')
 HOSTNAME = gethostname()
 IS_ADDON = bool(strtobool(getenv(f'{NAME}_IS_ADDON'.upper(), 'false')))
-
+PRESET_LAST_KEY = '.LAST'
 
 @dataclass
 class Config:
@@ -32,6 +34,9 @@ class Config:
     device_names: dict = None
     logging: str = None
     tele_period: int = 300
+    presets: dict = field(default_factory=dict)
+    config_raw: dict = field(default_factory=dict)
+
 
     def __post_init__(self):
         path_audio = Path(self.path_audio).absolute()
@@ -43,8 +48,13 @@ class Config:
         self.logging = self.logging or logging.INFO
 
     @classmethod
-    def from_file(cls):
+    @lru_cache
+    def get_path_config(cls) -> Path:
+        """
 
+        Get path to config file from environment variable, or default location
+
+        """
         path_config = getenv('AMNIOTIC_CONFIG_PATH')
 
         if not path_config:
@@ -53,6 +63,13 @@ class Config:
 
         path_config = Path(path_config).absolute()
 
+        return path_config
+
+    @classmethod
+    def from_file(cls):
+
+        path_config = cls.get_path_config()
+
         if not path_config.exists():
             msg = f'Config file not found at "{path_config}". Default values will be used.'
             logging.warning(msg)
@@ -60,18 +77,10 @@ class Config:
         else:
             msg = f'Config file found at "{path_config}"'
             logging.info(msg)
-
             config_str = Path(path_config).read_text()
+            config = yaml.safe_load(config_str)
 
-            if path_config.suffix in {'.yml', '.yaml'}:
-                config = yaml.safe_load(config_str)
-            elif path_config.suffix in {'.json'}:
-                config = json.loads(config_str)
-            else:
-                msg = f'Unknown config format "{path_config.suffix}"'
-                raise ValueError(msg)
-
-            logging.warning(msg)
+        config['config_raw'] = deepcopy(config)
 
         field_names = {field.name for field in fields(Config)}
         for key in field_names:
@@ -86,3 +95,30 @@ class Config:
 
         config = cls(**config)
         return config
+
+    def write(self) -> int:
+        """
+
+        Write out current config to same path from where it was read
+
+        """
+        config_str = yaml.dump(self.config_raw)
+        path = self.get_path_config()
+        logging.info(f'Wrote out config file to: {path}')
+
+        return path.write_text(config_str)
+
+    def write_presets(self, presets: Dict, preset_last: Optional[Dict] = None) -> int:
+        """
+
+        Write out current config to same path from where it was read
+
+        """
+
+        presets = deepcopy(presets)  # Don't modify the original dictionary
+        if preset_last:
+            presets[PRESET_LAST_KEY] = preset_last
+
+        self.config_raw['presets'] = presets
+
+        return self.write()
