@@ -1,3 +1,4 @@
+import json
 import logging
 import pip
 import threading
@@ -27,8 +28,11 @@ class Entity:
     HA_PLATFORM = None
     NAME = None
     ICON_SUFFIX = None
+    NA_VALUE = '-'
     value = None
 
+    TEST_ALWAYS_UPDATE = False
+
     def __init__(self, loop: Loop):
         self.loop = loop
 
@@ -172,8 +176,7 @@ class Entity:
 
         value = self.get_value()
         if value != self.value or force_announce:
-            self.set_value(value)
-            self.value = self.get_value()
+            self.value = value
             message = Message(self.client.publish, self.topic_state, self.value)
             self.queue.append(message)
 
@@ -242,10 +245,11 @@ class SelectTheme(Select):
     def get_options(self, amniotic: Amniotic) -> list[str]:
         """
 
-        Get state of the entity, i.e. the list of options and the currently selected option.
+        Ensure Themes are up-to-date with what's on disk, and return a list of their names
 
         """
-        return list(amniotic.themes.keys())
+        self.amniotic.load_themes()
+        return sorted(amniotic.themes.keys())
 
 
 class Volume(Entity):
@@ -375,7 +379,7 @@ class ToggleTheme(Entity):
 class Button(Entity):
     """
 
-    Home Assistant button.
+    Home Assistant button
 
     """
     HA_PLATFORM = 'button'
@@ -387,9 +391,27 @@ class Button(Entity):
         pass
 
 
-class ButtonVolumeDown(Button):
+class ButtonDisableAllThemes(Button):
+    """
+
+    Home Assistant button to disable all Themes
+
+    """
+    NAME = 'Disable All Themes'
+    ICON_SUFFIX = 'stop-circle'
+
+    def handle_incoming(self, value: Any):
+        """
+
+        Disable all themes
+
+        """
+        self.amniotic.enabled = False
+
+
+class ButtonVolumeDownMaster(Button):
     NAME = 'Master Volume Down'
-    ICON_SUFFIX = 'volume-low'
+    ICON_SUFFIX = 'volume-minus'
 
     def handle_incoming(self, value: Any):
         """
@@ -400,9 +422,9 @@ class ButtonVolumeDown(Button):
         self.amniotic.set_volume_down()
 
 
-class ButtonVolumeUp(Button):
+class ButtonVolumeUpMaster(Button):
     NAME = 'Master Volume Up'
-    ICON_SUFFIX = 'volume-high'
+    ICON_SUFFIX = 'volume-plus'
 
     def handle_incoming(self, value: Any):
         """
@@ -413,6 +435,45 @@ class ButtonVolumeUp(Button):
         self.amniotic.set_volume_up()
 
 
+class ButtonVolumeDownTheme(Button):
+    NAME = 'Theme Volume Down'
+    ICON_SUFFIX = 'volume-minus'
+
+    def handle_incoming(self, value: Any):
+        """
+
+        Decrement volume
+
+        """
+        self.amniotic.theme_current.set_volume_down()
+
+
+class ButtonVolumeUpTheme(Button):
+    NAME = 'Theme Volume Up'
+    ICON_SUFFIX = 'volume-plus'
+
+    def handle_incoming(self, value: Any):
+        """
+
+        Increment volume
+
+        """
+        self.amniotic.theme_current.set_volume_up()
+
+
+class ButtonRestart(Button):
+    NAME = 'Restart'
+    ICON_SUFFIX = 'restart-alert'
+
+    def handle_incoming(self, value: Any):
+        """
+
+        Tell Loop to end
+
+        """
+        self.loop.exit_reason = 'Restart requested'
+
+
 class ButtonUpdateCheck(Button):
     """
 
@@ -532,8 +593,7 @@ class ButtonUpdate(ButtonUpdateCheck):
 class TextInput(Entity):
     """
 
-    Base Home Assistant text input box. Note: this control abuses an alarm panel code entry box, as it seems to be the only way to allow a user to send
-    arbitrary text (e.g. a URL) from a Home Assistant control.
+    Base Home Assistant text input box.
 
     """
     HA_PLATFORM = 'text'
@@ -559,7 +619,7 @@ class TextInput(Entity):
     def get_value(self) -> Any:
         """
 
-        The current state of this control. Pending means Idle, and Triggered means Downloading.
+        The current state of this control.
 
         """
         return self.status
@@ -613,8 +673,8 @@ class Downloader(TextInput):
         Get the sensor for displaying update messages
 
         """
-        from amniotic.mqtt.sensor import DownloaderStatus
-        update_status = self.loop.entities[DownloaderStatus]
+        from amniotic.mqtt.sensor import DownloadStatus
+        update_status = self.loop.entities[DownloadStatus]
         return update_status
 
     def progress_callback(self, stream: Stream, chunk: bytes, bytes_remaining: int):
@@ -645,7 +705,6 @@ class Downloader(TextInput):
         """
 
         try:
-
             self.status = self.DOWNLOADING
             theme = self.amniotic.theme_current
             self.update_sensor.message = 'Fetching video metadata...'
@@ -688,3 +747,158 @@ class Downloader(TextInput):
         if self.status == self.DOWNLOADING:
             return
         threading.Thread(target=self.do_download, args=[value]).start()
+
+
+class PresetData(TextInput):
+    """
+
+    Home Assistant text input box for I/O of Presets
+
+    """
+    ICON_SUFFIX = 'cog-play'
+    NAME = 'Preset Data'
+
+    def update_sensor(self):
+        pass
+
+    def set_value(self, value) -> Any:
+        """
+
+        Apply preset data
+
+        """
+        self.amniotic.apply_preset_data(value)
+
+    def get_value(self) -> str:
+        """
+
+        Current settings as Preset data
+
+        """
+        preset = self.amniotic.get_preset_data()
+        preset_json = json.dumps(preset)
+        return preset_json
+
+
+class SaveDeletePreset(TextInput):
+    """
+
+    Home Assistant base text input box for saving/deleting Presets
+
+    """
+    PATTERN = '^[^\.].*$'
+
+    def update_sensor(self):
+        pass
+
+    @property
+    def data(self):
+        """
+
+        Home Assistant announce data for the entity. Don't allow Preset names to start with a '.' because these will be
+        used internally as special cases (e.g. last-used Preset)
+
+        """
+        data = super().data | {
+            # 'pattern': self.PATTERN TODO: Fix this
+        }
+        return data
+
+    def get_value(self) -> Optional[str]:
+        """
+
+        We can leave the text as the last value the user entered, so no need to return anything here
+
+        """
+        return None
+
+    def write_presets(self) -> int:
+        """
+
+        Write out presets after change
+
+        """
+        return self.loop.config.write_presets(
+            presets=self.amniotic.presets,
+            preset_last=self.amniotic.get_preset_data()
+        )
+
+
+class SavePreset(SaveDeletePreset):
+    """
+
+    Home Assistant text input box for naming/saving a new Preset
+
+    """
+    ICON_SUFFIX = 'content-save-plus'
+    NAME = 'Save Preset As'
+
+    def set_value(self, value):
+        """
+
+        Add a new Preset
+
+        """
+        self.amniotic.add_preset(value)
+        self.write_presets()
+
+
+class DeletePreset(SaveDeletePreset):
+    """
+
+    Home Assistant text input box for deleting Preset
+
+    """
+    ICON_SUFFIX = 'trash-can'
+    NAME = 'Delete Preset'
+    PATTERN = '^[^\.].*$'  # TODO: Inherit this (and data/update_sensor) from common "PresetIO".
+
+    def update_sensor(self):
+        pass
+
+    def set_value(self, value):
+        """
+
+        Remove a Preset
+
+        """
+        self.amniotic.remove_preset(value)
+        self.write_presets()
+
+
+class Preset(Select):
+    """
+
+    Home Assistant Presets selector.
+
+    """
+    ICON_SUFFIX = 'window-shutter-cog'
+    NAME = 'Preset'
+    NA_VALUE = ""
+
+    def get_value(self) -> str:
+        """
+
+        Get the currently selected Preset, if there is one
+
+        """
+        return self.amniotic.get_preset() or self.NA_VALUE
+
+    def set_value(self, value) -> str:
+        """
+
+        If selector isn't empty, apply the selected Preset
+
+        """
+        if value == self.NA_VALUE:
+            return
+        self.amniotic.apply_preset(value)
+
+    def get_options(self, amniotic: Amniotic) -> list[str]:
+        """
+
+        Get state of the entity, i.e. the list of options and the currently selected option. Also allow no entry to
+        be selected
+
+        """
+        return [self.NA_VALUE] + sorted(amniotic.presets.keys())
